#include <malloc.h>
#include "allocator.h"

static void* STDLib_allocator_malloc(Allocator _this, size_t size);
static void* STDLib_allocator_realloc(Allocator _this, void* ptr, size_t size);
static void  STDLib_allocator_free(Allocator _this, void* ptr);

static void  STDLib_allocator_releaseAll(Allocator _this);

static vt_allocator_t VT_LinkedSampleAllocator = {
    STDLib_allocator_malloc,
    STDLib_allocator_realloc,
    STDLib_allocator_free,

    STDLib_allocator_releaseAll,
};

typedef struct _LinkedSampleAllocator  LinkedSampleAllocator_t, *LinkedSampleAllocator_p_t;

typedef struct _block_header block_header_t, *block_header_p_t;
struct _block_header {
    block_header_p_t           next;
    block_header_p_t           prev;
};

struct _LinkedSampleAllocator {
    vt_allocator_p_t vtable;
    block_header_t   m_rootBlock;
};

#define this  ((LinkedSampleAllocator_p_t)_this)

Allocator createLocalAllocator()
{
    Allocator _this = malloc(sizeof(LinkedSampleAllocator_t));
    if (!_this) return 0;
    this->vtable = &VT_LinkedSampleAllocator;
    this->m_rootBlock.next = this->m_rootBlock.prev = &this->m_rootBlock;
    return _this;
}


static void* STDLib_allocator_malloc(Allocator _this, size_t size)
{
    block_header_p_t result = malloc(size+sizeof(block_header_t));
    if (!result) {
        return 0;
        /* throwException(&class_OutOfMemoryError, 0, 0); */
    }
    result->next = (result->prev = &this->m_rootBlock)->next;
    result->prev->next = result->next->prev = result;

    return result+1;
}

static void STDLib_allocator_free(Allocator _this, void* ptr)
{
    block_header_p_t block = (block_header_p_t)ptr-1;
    (block->next->prev = block->prev)->next = block->next;
    free(block);
}

static void* STDLib_allocator_realloc(Allocator _this, void* ptr, size_t size)
{
    if (!ptr) {
        return (*_this)->malloc(_this, size);
    } else if (!size) {
        (*_this)->free(_this, ptr);
        return 0;
    } else {
        block_header_p_t block  = (block_header_p_t)ptr-1;
        block_header_p_t result = realloc(block, size+sizeof(block_header_t));

        if (result && result != block) {
            result->next->prev = result->prev->next = result;
        }
        if (!result) {
            return 0;
        }
        return result+1;
    }
}

static void   STDLib_allocator_releaseAll(Allocator _this)
{
    block_header_p_t root = &this->m_rootBlock;
    while (root->next != root) {
        (*_this)->free(_this, root->next+1);
    }
}

